/*
 * A mini transition relation postfix reader
 * ---
 * January 2002
 * to my angel
 */

#include "checker.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "bdd.h"
#include "vars.h"
#include "mem.h"
#include "trans.h"
#include "ctl.h"
#include "verbose.h"

void print_tree(trans_data *d, variables *v, transition_tree t)
{
  static int current = -1;
  int pcur = current;

  if (trans_undef(t)) {
    fprintf(stdout, "(transition tree undefined)");
    return;
  }

  current = get_transition_tree_type(d, t);

  switch (current) {
    case VAR :
      fprintf(stdout, "%s", get_variable_name(v, get_transition_tree_variable(d, t)));
      break;
    case CTL_TRUE :
      fprintf(stdout, "true");
      break;
    case CTL_FALSE :
      fprintf(stdout, "false");
      break;
    case OR :
    case XOR :
      if (pcur == AND || pcur == NOT)
        fprintf(stdout, "(");
      print_tree(d, v, get_transition_tree_left_son(d, t));
      fprintf(stdout, "%c", current == OR ? '+' : '^');
      print_tree(d, v, get_transition_tree_right_son(d, t));
      if (pcur == AND || pcur == NOT)
        fprintf(stdout, ")");
      break;
    case AND :
      if (pcur == NOT)
        fprintf(stdout, "(");
      print_tree(d, v, get_transition_tree_left_son(d, t));
      fprintf(stdout, ".");
      print_tree(d, v, get_transition_tree_right_son(d, t));
      if (pcur == NOT)
        fprintf(stdout, ")");
      break;
    case IMPLIES :
      if (pcur != -1)
        fprintf(stdout, "(");
      print_tree(d, v, get_transition_tree_left_son(d, t));
      fprintf(stdout, "=>");
      print_tree(d, v, get_transition_tree_right_son(d, t));
      if (pcur != -1)
        fprintf(stdout, ")");
      break;
    case EQUAL :
      if (pcur != -1)
        fprintf(stdout, "(");
      print_tree(d, v, get_transition_tree_left_son(d, t));
      fprintf(stdout, "=");
      print_tree(d, v, get_transition_tree_right_son(d, t));
      if (pcur != -1)
        fprintf(stdout, ")");
      break;
    case NOT :
      fprintf(stdout, "!");
      print_tree(d, v, get_transition_tree_son(d, t));
      break;
    default :
      fprintf(stdout, "(unknown type %d)", current);
      break;
  }

  current = pcur;
}

int init_checker(checker *c, char *fname, int we_do)
{
  FILE *f = fopen(fname, "r");
  int ret;

  if (!f) {
    perror(fname);
    return -1;
  }

  memset(c, 0, sizeof(checker));

  c->filename = fname;
  c->we_do = we_do;

  if (init_vars(&c->variables) == -1) {
    close_checker(c);
    goto error;
  }

  if (we_do != WE_DO_BMC)
    if (init_bdd(&c->bdd_data) == -1) {
      close_checker(c);
      goto error;
    }

  if (init_trans(&c->trans_data) == -1) {
    close_checker(c);
    goto error;
  }

  if (init_ctl(&c->ctl_data) == -1) {
    close_checker(c);
    goto error;
  }

#if 0
  fsm_restart(f);
  fsm_linenumber = 1;
  ret = fsm_parse();
  fsm_lexer_close();
#endif

  if (ret == 1) {
    close_checker(c);
    goto error;
  }
  if (we_do != WE_DO_BMC)
    if (!c->init_states || !c->f) {
      close_checker(c);
      goto error;
    }

  fclose(f);
  return 0;

error:
  fclose(f);
  return -1;
}

void close_checker(checker *c)
{
  if (c->vars)
    free_bdd_variables_set(&c->bdd_data, c->vars);
  if (c->varsnext)
    free_bdd_variables_set(&c->bdd_data, c->varsnext);
  if (c->varsinput)
    free_bdd_variables_set(&c->bdd_data, c->varsinput);
  if (c->varsoutput)
    free_bdd_variables_set(&c->bdd_data, c->varsoutput);
  if (c->we_do != WE_DO_BMC)
    free_bdd(&c->bdd_data);
  free_vars(&c->variables);
  free_trans(&c->trans_data);
  free_ctl(&c->ctl_data);

  if (c->nbmax_fairness)
    freemem(c->fairness);
  if (c->nb_fairness && c->fairness_bdd)
    freemem(c->fairness_bdd);

  memset(c, 0, sizeof(checker));
}

static bdd *do_checkEX(checker *c, bdd *a)
{
  bdd *ret=bdd_get_predecessors(&c->bdd_data, c->f, a, c->vars, c->varsnext, c->varsinput);
  crash(ret);
  return ret;
}

static bdd *do_checkEG(checker *c, bdd *a)
{
  bdd *Q, *Q_;

  Q = bdd_one(&c->bdd_data);
  crash(Q);
  Q_ = a;

  while (Q != Q_) {
    Q = Q_;
    Q_ = bdd_and(&c->bdd_data, a, do_checkEX(c, Q_));
    crash(Q_);
  }

  return Q;
}

static bdd *do_checkEU(checker *c, bdd *a, bdd *b)
{
  bdd *Q, *Q_;

  Q = bdd_zero(&c->bdd_data);
  crash(Q);
  Q_ = b;

  while (Q != Q_) {
    Q = Q_;
    Q_ = bdd_and(&c->bdd_data, a, do_checkEX(c, Q_));
    crash(Q_);
    Q_ = bdd_or(&c->bdd_data, b, Q_);
    crash(Q_);
  }

  return Q;
}

/* forward declaration */
static bdd *do_check(checker *, ctl_formula);

static bdd *do_checkE(checker *c, ctl_formula f)
{
  bdd *a, *b;

  switch(get_ctl_type(&c->ctl_data, f)) {
    case NEXT :
      a = do_check(c, get_ctl_son(&c->ctl_data, f));
      crash(a);
      return do_checkEX(c, a);
    case ALWAYS :
      a = do_check(c, get_ctl_son(&c->ctl_data, f));
      crash(a);
      return do_checkEG(c, a);
    case UNTIL :
      a = do_check(c, get_ctl_left_son(&c->ctl_data, f));
      crash(a);
      b = do_check(c, get_ctl_right_son(&c->ctl_data, f));
      crash(b);
      return do_checkEU(c, a, b);
    default :
      fprintf(stderr, "checker.c (line %d) do_checkE: error: impossible case\n", __LINE__);
      abort();
  }
}

static bdd *do_check(checker *c, ctl_formula f)
{
  bdd *a, *b;

  switch(get_ctl_type(&c->ctl_data, f)) {
    case ATOM :
      return (bdd *)get_variable_assoc(&c->variables, get_ctl_variable(&c->ctl_data, f));
    case ONE_PATH :
      a = do_checkE(c, get_ctl_son(&c->ctl_data, f));
      crash(a);
      return a;
    case AND :
      a = do_check(c, get_ctl_left_son(&c->ctl_data, f));
      b = do_check(c, get_ctl_right_son(&c->ctl_data, f));
      a = bdd_and(&c->bdd_data, a, b);
      crash(a);
      return a;
    case OR :
      a = do_check(c, get_ctl_left_son(&c->ctl_data, f));
      b = do_check(c, get_ctl_right_son(&c->ctl_data, f));
      a = bdd_or(&c->bdd_data, a, b);
      crash(a);
      return a;
    case XOR :
      a = do_check(c, get_ctl_left_son(&c->ctl_data, f));
      b = do_check(c, get_ctl_right_son(&c->ctl_data, f));
      a = bdd_xor(&c->bdd_data, a, b);
      crash(a);
      return a;
    case NOT :
      a = do_check(c, get_ctl_son(&c->ctl_data, f));
      a = bdd_not(&c->bdd_data, a);
      crash(a);
      return a;
    case IMPLIES :
      a = do_check(c, get_ctl_left_son(&c->ctl_data, f));
      a = bdd_not(&c->bdd_data, a);
      crash(a);
      b = do_check(c, get_ctl_right_son(&c->ctl_data, f));
      a = bdd_or(&c->bdd_data, a, b);
      crash(a);
      return a;
    case CTL_TRUE :
      a = bdd_one(&c->bdd_data);
      crash(a);
      return a;
    case CTL_FALSE :
      a = bdd_zero(&c->bdd_data);
      crash(a);
      return a;
    default :
      fprintf(stderr, "checker.c (line %d) do_check: error: impossible case\n", __LINE__);
      abort();
  }
}

static bdd *do_checkfairEX(checker *c, bdd *a, bdd *fair)
{
  bdd *aa = bdd_and(&c->bdd_data, a, fair);
  crash(aa);
  return do_checkEX(c, aa);
}

static bdd *do_checkfairEU(checker *c, bdd *a, bdd *b, bdd *fair)
{
  bdd *bb = bdd_and(&c->bdd_data, b, fair);
  crash(bb);
  return do_checkEU(c, a, bb);
}

inline static bdd *TauEG(bdd *Q, checker *c, bdd *f)
{
  int i;
  bdd *P = c->fairness_bdd[0];
  bdd *Z = bdd_and(&c->bdd_data, Q, P);

  crash(Z);
  Z = do_checkEU(c, f, Z);
  Z = do_checkEX(c, Z);

  for (i=1; i<c->nb_fairness; i++) {
    bdd *Z1;

    P = c->fairness_bdd[i];

    Z1 = bdd_and(&c->bdd_data, Q, P);
    crash(Z1);

    Z1 = do_checkEU(c, f, Z1);
    Z1 = do_checkEX(c, Z1);

    Z = bdd_and(&c->bdd_data, Z, Z1);
    crash(Z);
  }

  Z = bdd_and(&c->bdd_data, f, Z);
  crash(Z);

  return Z;
}

static bdd *do_checkfairEG(checker *c, bdd *f)
{
  bdd *Q, *Q_;

  /* a greatest fixpoint */
  Q = bdd_one(&c->bdd_data);
  Q_ = TauEG(Q, c, f);

  while (Q != Q_) {
    Q = Q_;
    Q_ = TauEG(Q, c, f);
  }

  return Q;
}

/* forward declaration */
static bdd *do_checkfair_(checker *, ctl_formula, bdd *);

static bdd *do_checkfairE(checker *c, ctl_formula f, bdd *fair)
{
  bdd *a, *b;

  switch(get_ctl_type(&c->ctl_data, f)) {
    case NEXT :
      a = do_checkfair_(c, get_ctl_son(&c->ctl_data, f), fair);
      crash(a);
      return do_checkfairEX(c, a, fair);
    case ALWAYS :
      a = do_checkfair_(c, get_ctl_son(&c->ctl_data, f), fair);
      crash(a);
      return do_checkfairEG(c, a);
    case UNTIL :
      a = do_checkfair_(c, get_ctl_left_son(&c->ctl_data, f), fair);
      crash(a);
      b = do_checkfair_(c, get_ctl_right_son(&c->ctl_data, f), fair);
      crash(b);
      return do_checkfairEU(c, a, b, fair);
    default :
      fprintf(stderr, "checker.c (line %d) do_checkE: error: impossible case\n", __LINE__);
      abort();
  }
}

static bdd *do_checkfair_(checker *c, ctl_formula f, bdd *fair)
{
  bdd *a, *b;

  switch(get_ctl_type(&c->ctl_data, f)) {
    case ATOM :
      return (bdd *)get_variable_assoc(&c->variables, get_ctl_variable(&c->ctl_data, f));
    case ONE_PATH :
      a = do_checkfairE(c, get_ctl_son(&c->ctl_data, f), fair);
      crash(a);
      return a;
    case AND :
      a = do_checkfair_(c, get_ctl_left_son(&c->ctl_data, f), fair);
      b = do_checkfair_(c, get_ctl_right_son(&c->ctl_data, f), fair);
      a = bdd_and(&c->bdd_data, a, b);
      crash(a);
      return a;
    case OR :
      a = do_checkfair_(c, get_ctl_left_son(&c->ctl_data, f), fair);
      b = do_checkfair_(c, get_ctl_right_son(&c->ctl_data, f), fair);
      a = bdd_or(&c->bdd_data, a, b);
      crash(a);
      return a;
    case XOR :
      a = do_checkfair_(c, get_ctl_left_son(&c->ctl_data, f), fair);
      b = do_checkfair_(c, get_ctl_right_son(&c->ctl_data, f), fair);
      a = bdd_xor(&c->bdd_data, a, b);
      crash(a);
      return a;
    case NOT :
      a = do_checkfair_(c, get_ctl_son(&c->ctl_data, f), fair);
      a = bdd_not(&c->bdd_data, a);
      crash(a);
      return a;
    case IMPLIES :
      a = do_checkfair_(c, get_ctl_left_son(&c->ctl_data, f), fair);
      a = bdd_not(&c->bdd_data, a);
      crash(a);
      b = do_checkfair_(c, get_ctl_right_son(&c->ctl_data, f), fair);
      a = bdd_or(&c->bdd_data, a, b);
      crash(a);
      return a;
    case CTL_TRUE :
      a = bdd_one(&c->bdd_data);
      crash(a);
      return a;
    case CTL_FALSE :
      a = bdd_zero(&c->bdd_data);
      crash(a);
      return a;
    default :
      fprintf(stderr, "checker.c (line %d) do_checkfair_: error: impossible case\n", __LINE__);
      abort();
  }
}

static bdd *do_checkfair(checker *c, ctl_formula f)
{
  bdd *fair;

  if (c->nb_fairness <= 0) {
    print("check", stdout, "checker.c:do_checkfair:%d: warning : calling do_checkfair with 0 fairness constraint, running do_check instead\n", __LINE__);

    return do_check(c, f);
  }

  fair = do_checkfairEG(c, bdd_one(&c->bdd_data));

  return do_checkfair_(c, f, fair);
}

static bdd **list = 0;
static int maxsize = 0, size = 0;

void add_trace(bdd *b)
{
  if (size == maxsize)
    list = regetmem(list, sizeof(bdd *)*(maxsize+=16));

  list[size++] = b;
}

/* this function prints a trace from init state to a state
 * not verifying the property (ie not in d)
 */
void trace(checker *c, bdd *d)
{
  bdd *cur;
  /* graph is the accessible state graph, we calculate it on the fly */
  bdd *graph;
  int i;

  d = bdd_not(&c->bdd_data, d);

  print("check_trace", stdout, "Here comes the bdd representing all the states NOT verifiying the property\n");
  if (verbose_on("check_trace"))
    bdd_print(&c->bdd_data, &c->variables, d, BDD_PRINT_TEXT);

  graph = c->init_states;

  cur = c->init_states;
  add_trace(cur);

  print("check_trace", stdout, "first trace ::::::::::::::::::\n");
  if (verbose_on("check_trace"))
    bdd_print(&c->bdd_data, &c->variables, cur, BDD_PRINT_TEXT);

  while (bdd_is_zero(&c->bdd_data, bdd_and(&c->bdd_data, cur, d))) {
    bdd *tmp=graph;
    cur = bdd_get_successors(&c->bdd_data, c->f, cur, c->vars, c->varsnext, c->varsinput);
    graph = bdd_or(&c->bdd_data, cur, graph);
    if (graph == tmp) {
      print("check_trace", stdout, "No trace exist, no bad state is reachable\n");
      freemem(list);
      list =0;
      size = maxsize = 0;
      return;
    }
    add_trace(cur);
    print("check_trace", stdout, "next trace ::::::::::::::::::\n");
    if (verbose_on("check_trace"))
      bdd_print(&c->bdd_data, &c->variables, cur, BDD_PRINT_TEXT);
  }

  i = size-1;

  list[i] = bdd_and(&c->bdd_data, list[i], d);

  while (i) {
    cur = bdd_get_one_state(&c->bdd_data, list[i], c->vars->list, c->variables.size);
    list[i] = cur;
    cur = bdd_get_predecessors(&c->bdd_data, c->f, cur, c->vars, c->varsnext, c->varsinput);
    list[i-1] = bdd_and(&c->bdd_data, list[i-1], cur);
    i--;
  }
  list[0] = bdd_get_one_state(&c->bdd_data, list[0], c->vars->list, c->variables.size);

  print("check_trace", stdout, "the trace has %d nodes\n", size);
  if (verbose_on("check_trace"))
    for (i = 0; i<size; i++)
      bdd_print(&c->bdd_data, &c->variables, list[i], BDD_PRINT_TEXT);

  if (list)
    freemem(list);
  list = 0;
  maxsize = size = 0;
}

/* return 0 if the formula is false, 1 if it is true on our system, -1 if error */
int check(checker *c)
{
  bdd *d, *res;
  ctl_formula f = c->formula;

  c->formula = normalize_ctl_formula(&c->ctl_data, f);
  free_ctl_formula(&c->ctl_data, f);

  print("ctl_parser", stdout, "here is the ctl formula that will be checked :\n");
  if (verbose_on("ctl_parser"))
    print_formula(&c->ctl_data, &c->variables, c->formula);
  print("ctl_parser", stdout, "\n");

  if (c->nb_fairness) {
    int i;
    print("check", stdout, "calling do_checkfair\n");

    /* let's first calculate the bdd of the fairness formulas */
    c->fairness_bdd = getmem(c->nb_fairness * sizeof(bdd *));
    for (i=0; i<c->nb_fairness; i++) {
      ctl_formula ff = c->fairness[i];
      c->fairness[i] = normalize_ctl_formula(&c->ctl_data, ff);
      free_ctl_formula(&c->ctl_data, f);
      c->fairness_bdd[i] = do_check(c, c->fairness[i]);

      print("ctl_parser", stdout, "constraint ctl formula : ");
      if (verbose_on("ctl_parser"))
	print_formula(&c->ctl_data, &c->variables, c->fairness[i]);
      print("ctl_parser", stdout, "\n");
      print("ctl_parser", stdout, "the states ok for this constraint :\n");
      if (verbose_on("ctl_parser"))
	bdd_print(&c->bdd_data, &c->variables, c->fairness_bdd[i], BDD_PRINT_TEXT);
    }
    /* and now let's check our property */
    d = do_checkfair(c, c->formula);
  } else {
    print("check", stdout, "calling do_check\n");

    d = do_check(c, c->formula);
  }

  print("check_bdd", stdout, "Here comes the bdd representing all the states verifiying the property\n");
  if (verbose_on("check_bdd"))
    bdd_print(&c->bdd_data, &c->variables, d, BDD_PRINT_TEXT);

  trace(c, d);

  res = bdd_and(&c->bdd_data, d, c->init_states);
  crash(res);
  return !bdd_is_zero(&c->bdd_data, res);
}
