/*
 * A CTL parser.
 * The ctl tree definition.
 *
 *----
 * To my angel,
 * march 2002.
 *----
 * This file is in the public domain.
 */

#include "ctl.h"

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>

#include "mem.h"
#include "vars.h"
#include "trans.h"
#include "hash.h"
#include "verbose.h"

#define inc(x) if ((x) < UINT_MAX-1) (x) = (x)+1
#define dec(x) if ((x) != 0) (x) = (x) - 1

#define incref(d, f) inc((d)->tab[f].nb_refs)
#define decref(d, f) dec((d)->tab[f].nb_refs)

static int ctl_compare(ctl_data *d, _ctl_formula *new, ctl_formula *in_hash)
{
  _ctl_formula *old = &d->tab[*in_hash];

  /* nb_refs makes the difference, so we can't use memcmp */
  if (old->op != new->op)
    return 0;

  switch(old->op) {
    case CTL_TRUE :
    case CTL_FALSE :
      /* true and false only exist in op, so here we know they are == */
      return 1;
    case ATOM :
      return old->val.variable == new->val.variable;
    case ALL_PATH :
    case ONE_PATH :
    case ALWAYS :
    case EVENTUALLY :
    case NEXT :
    case NOT :
      return old->val.op_one.s == new->val.op_one.s;
    case UNTIL :
    case AND :
    case OR :
    case XOR :
    case IMPLIES :
      return old->val.op_two.l == new->val.op_two.l
          && old->val.op_two.r == new->val.op_two.r;
  }

  /* we cannot come here */
  fprintf(stderr, "ctl.c:%d:ctl_compare : error, unreachable code reached\n", __LINE__);
  abort();

  /* we want no compiler complains */
  return 0;
}

static unsigned long ctl_hash(ctl_data *d, _ctl_formula *f)
{
  return hash_simple(&d->ht, f, sizeof(_ctl_formula));
}

static _ctl_formula *ctl_get_key(ctl_data *d, ctl_formula *f)
{
  return &d->tab[*f];
}

/* return 0 if ok, -1 if not */
int init_ctl(ctl_data *d)
{
  memset(d, 0, sizeof(ctl_data));

  init_hash(&d->ht, sizeof(ctl_formula), HASH_INIT_SIZE_SMALL,
	    (hash_compare_function)ctl_compare, d,
            (hash_hash_function)ctl_hash, d,
            (hash_get_key_function)ctl_get_key, d);

  return 0;
}

void free_ctl(ctl_data *d)
{
  if (d->maxsize && d->tab)
    freemem(d->tab);

  d->tab = 0;
  d->size = d->maxsize = 0;

  print("hash", stdout, "ctl.c:free_ctl : ");
  free_hash(&d->ht);
}

void free_ctl_formula(ctl_data *d, ctl_formula f)
{
  switch (get_ctl_type(d, f)) {
    case ATOM :
    case CTL_TRUE :
    case CTL_FALSE :
      decref(d, f);
      return;
    case ALL_PATH :
    case ONE_PATH :
    case ALWAYS :
    case EVENTUALLY :
    case NEXT :
    case NOT :
      free_ctl_formula(d, get_ctl_son(d, f));
      decref(d, f);
      return;
    case AND :
    case OR :
    case XOR :
    case IMPLIES :
    case UNTIL :
      free_ctl_formula(d, get_ctl_left_son(d, f));
      free_ctl_formula(d, get_ctl_right_son(d, f));
      decref(d, f);
      return;
    default :
      /* one can't come here */
      fprintf(stderr, "ctl.c:%d:free_ctl_formula : error : unreachable code reached\n", __LINE__);
      abort();
  }
}

ctl_formula new_ctl_formula(ctl_data *d)
{
  if (d->size == d->maxsize)
    d->tab = regetmem(d->tab, (d->maxsize+=128) * sizeof(_ctl_formula));

  memset(&d->tab[d->size], 0, sizeof(_ctl_formula));

  incref(d, d->size);

  d->size++;

  return d->size-1;
}

#define clear_ctl_formula(f) memset(&(f), 0, sizeof(_ctl_formula))

ctl_formula new_ctl_formula_var(ctl_data *d, variable v)
{
  _ctl_formula _f;
  ctl_formula f;
  hid h;

  /* handle hash table */
  clear_ctl_formula(_f);
  _f.op = ATOM;
  _f.val.variable = v;

  h = hash_get_new_slot(&d->ht, &_f);
  if (d->ht.flag == HASH_ALREADY_PRESENT) {
    f = *(ctl_formula *)hash_get_slot_data(&d->ht, h);
    incref(d, f);
    print("hash", stdout, "hash already present : ctl node %u = var %u\n", f, v);
    return f;
  }

  /* handle data */
  f = new_ctl_formula(d);

  hash_feed_slot(&d->ht, h, &f);

  set_ctl_type(d, f, ATOM);
  set_ctl_variable(d, f, v);

  return f;
}

ctl_formula new_ctl_formula_op(ctl_data *d, int op, ctl_formula l, ctl_formula r)
{
  _ctl_formula _f;
  ctl_formula f;
  hid h;

  /* sanity checks */
  switch (op) {
    case OR :
    case XOR :
    case AND :
    case EQUAL :
    case UNTIL :
    case IMPLIES :
      if (l == CTL_NO_FORMULA || r == CTL_NO_FORMULA) {
	fprintf(stderr, "ctl.c %d fatal error\n", __LINE__);
	abort();
      }
      break;
    case NOT :
    case ALL_PATH :
    case ONE_PATH :
    case ALWAYS :
    case EVENTUALLY :
    case NEXT :
      if (l == CTL_NO_FORMULA || r != CTL_NO_FORMULA) {
	fprintf(stderr, "ctl.c %d fatal error\n", __LINE__);
	abort();
      }
      break;
    default :
      fprintf(stderr, "ctl.c %d fatal error, unreachable code reached !?\n", __LINE__);
      abort();
  }

  /* handle hash table */
  clear_ctl_formula(_f);
  _f.op = op;
  switch (op) {
    case OR :
    case XOR :
    case AND :
    case EQUAL :
    case UNTIL :
    case IMPLIES :
      _f.val.op_two.l = l;
      _f.val.op_two.r = r;
      break;
    case NOT :
    case ALL_PATH :
    case ONE_PATH :
    case ALWAYS :
    case EVENTUALLY :
    case NEXT :
      _f.val.op_one.s = l;
      break;
  }

  h = hash_get_new_slot(&d->ht, &_f);
  if (d->ht.flag == HASH_ALREADY_PRESENT) {
    f = *(ctl_formula *)hash_get_slot_data(&d->ht, h);
    incref(d, f);
    print("hash", stdout, "hash already present : ctl node %u = op %u args %u %u\n", f, op, l, r);
    return f;
  }

  /* handle data */
  f = new_ctl_formula(d);

  hash_feed_slot(&d->ht, h, &f);

  set_ctl_type(d, f, op);

  switch(op) {
    case OR :
    case XOR :
    case AND :
    case EQUAL :
    case UNTIL :
    case IMPLIES :
      set_ctl_left_son(d, f, l);
      set_ctl_right_son(d, f, r);
      break;
    case NOT :
    case ALL_PATH :
    case ONE_PATH :
    case ALWAYS :
    case EVENTUALLY :
    case NEXT :
      set_ctl_son(d, f, l);
      break;
  }

  return f;
}

ctl_formula new_ctl_formula_true(ctl_data *d)
{
  _ctl_formula _f;
  ctl_formula f;
  hid h;

  /* handle hash table */
  clear_ctl_formula(_f);
  _f.op = CTL_TRUE;

  h = hash_get_new_slot(&d->ht, &_f);
  if (d->ht.flag == HASH_ALREADY_PRESENT) {
    f = *(ctl_formula *)hash_get_slot_data(&d->ht, h);
    incref(d, f);
    print("hash", stdout, "hash already present : ctl node %u = TRUE\n", f);
    return f;
  }

  /* handle data */
  f = new_ctl_formula(d);

  hash_feed_slot(&d->ht, h, &f);

  set_ctl_type(d, f, CTL_TRUE);
    
  return f;
}

ctl_formula new_ctl_formula_false(ctl_data *d)
{
  _ctl_formula _f;
  ctl_formula f;
  hid h;

  /* handle hash table */
  clear_ctl_formula(_f);
  _f.op = CTL_FALSE;

  h = hash_get_new_slot(&d->ht, &_f);
  if (d->ht.flag == HASH_ALREADY_PRESENT) {
    f = *(ctl_formula *)hash_get_slot_data(&d->ht, h);
    incref(d, f);
    print("hash", stdout, "hash already present : ctl node %u = FALSE\n", f);
    return f;
  }

  /* handle data */
  f = new_ctl_formula(d);

  hash_feed_slot(&d->ht, h, &f);

  set_ctl_type(d, f, CTL_FALSE);
    
  return f;
}

static ctl_formula all_path_normalize(ctl_data *d, ctl_formula f)
{
  switch (get_ctl_type(d, get_ctl_son(d, f))) {
    case NEXT : {
      /* AX a => !EX!norm(a) */
      ctl_formula a = get_ctl_son(d, get_ctl_son(d, f));
      a = normalize_ctl_formula(d, a);
      a = new_ctl_formula_op(d, NOT, a, CTL_NO_FORMULA);
      a = new_ctl_formula_op(d, NEXT, a, CTL_NO_FORMULA);
      a = new_ctl_formula_op(d, ONE_PATH, a, CTL_NO_FORMULA);
      return new_ctl_formula_op(d, NOT, a, CTL_NO_FORMULA);
    }
    case ALWAYS : {
      /* AG a => ! E(True U !norm(a)) */
      ctl_formula a = get_ctl_son(d, get_ctl_son(d, f));
      a = normalize_ctl_formula(d, a);
      a = new_ctl_formula_op(d, NOT, a, CTL_NO_FORMULA);
      a = new_ctl_formula_op(d, UNTIL, new_ctl_formula_true(d), a);
      a = new_ctl_formula_op(d, ONE_PATH, a, CTL_NO_FORMULA);
      return new_ctl_formula_op(d, NOT, a, CTL_NO_FORMULA);
    }
    case EVENTUALLY : {
      /* AF a => !EG!a */
      ctl_formula a = get_ctl_son(d, get_ctl_son(d, f));
      a = normalize_ctl_formula(d, a);
      a = new_ctl_formula_op(d, NOT, a, CTL_NO_FORMULA);
      a = new_ctl_formula_op(d, ALWAYS, a, CTL_NO_FORMULA);
      a = new_ctl_formula_op(d, ONE_PATH, a, CTL_NO_FORMULA);
      return new_ctl_formula_op(d, NOT, a, CTL_NO_FORMULA);
    }
    case UNTIL : {
      /* A(aUb) => !EG!b . !E(!b U !(a.b)) */
      ctl_formula a = get_ctl_left_son(d, get_ctl_son(d, f));
      ctl_formula b = get_ctl_right_son(d, get_ctl_son(d, f));
      register ctl_formula nota, notb, notEGnotb, notaandb, notEnotbUnotaandb;
      a = normalize_ctl_formula(d, a);
      b = normalize_ctl_formula(d, b);
      nota = new_ctl_formula_op(d, NOT, a, CTL_NO_FORMULA);
      notb = new_ctl_formula_op(d, NOT, b, CTL_NO_FORMULA);
      notEGnotb = new_ctl_formula_op(d, ALWAYS, notb, CTL_NO_FORMULA);
      notEGnotb = new_ctl_formula_op(d, ONE_PATH, notEGnotb, CTL_NO_FORMULA);
      notEGnotb = new_ctl_formula_op(d, NOT, notEGnotb, CTL_NO_FORMULA);
      notaandb = new_ctl_formula_op(d, AND, a, b);
      notaandb = new_ctl_formula_op(d, NOT, notaandb, CTL_NO_FORMULA);
      notEnotbUnotaandb = new_ctl_formula_op(d, UNTIL, notb, notaandb);
      notEnotbUnotaandb = new_ctl_formula_op(d, ONE_PATH, notEnotbUnotaandb, CTL_NO_FORMULA);
      notEnotbUnotaandb = new_ctl_formula_op(d, NOT, notEnotbUnotaandb, CTL_NO_FORMULA);
      return new_ctl_formula_op(d, AND, notEGnotb, notEnotbUnotaandb);
    }
    default :
      fprintf(stderr, "error in all_path_normalize, I can't come here ! byebye\n");
      abort();
  }
}

static ctl_formula one_path_normalize(ctl_data *d, ctl_formula f)
{
  switch (get_ctl_type(d, get_ctl_son(d, f))) {
    case EVENTUALLY : {
      /* EFa => E(TrueUa) */
      ctl_formula a = get_ctl_son(d, get_ctl_son(d, f));
      a = normalize_ctl_formula(d, a);
      a = new_ctl_formula_op(d, UNTIL, new_ctl_formula_true(d), a);
      return new_ctl_formula_op(d, ONE_PATH, a, CTL_NO_FORMULA);
    }
    case UNTIL : {
      /* E(aUb) => E(aUb) */
      ctl_formula a = get_ctl_left_son(d, get_ctl_son(d, f));
      ctl_formula b = get_ctl_right_son(d, get_ctl_son(d, f));
      register ctl_formula nf;
      ctl_formula na = normalize_ctl_formula(d, a);
      ctl_formula nb = normalize_ctl_formula(d, b);
      if (a == na && b == nb) {
	incref(d, f);
	return f;
      }
      nf = new_ctl_formula_op(d, UNTIL, na, nb);
      return new_ctl_formula_op(d, ONE_PATH, nf, CTL_NO_FORMULA);
    }
    case ALWAYS :
    case NEXT : {
      /* we return the same (unless normalization changed stuffs) */
      ctl_formula a = get_ctl_son(d, get_ctl_son(d, f));
      ctl_formula na = normalize_ctl_formula(d, a);
      if (a == na) {
	incref(d, f);
	return f;
      }
      na = new_ctl_formula_op(d, get_ctl_type(d, get_ctl_son(d, f)), na, CTL_NO_FORMULA);
      return new_ctl_formula_op(d, ONE_PATH, na, CTL_NO_FORMULA);
    }
    default :
      fprintf(stderr, "error in one_path_normalize, I can't come here ! byebye\n");
      abort();
  }
}

/* input : a ctl formula, output : the same with only  EX, EG and EU */
/* we copy, because of the use of hash table, one can't modify a ctl
 * formula wich would mean change its hash value, which
 * would mean take it away from hash table and back again,
 * which is a pain.
 * If you don't have the use of the previous formula, after this call
 * free_ctl_formula.
 */
ctl_formula normalize_ctl_formula(ctl_data *d, ctl_formula f)
{
  ctl_formula s1, s2, ns1, ns2;
  switch (get_ctl_type(d, f)) {
    case ATOM :
    case CTL_TRUE :
    case CTL_FALSE :
      incref(d, f);
      return f;
    case ALL_PATH :
      return all_path_normalize(d, f);
    case ONE_PATH :
      return one_path_normalize(d, f);
    case AND :
    case OR :
    case XOR :
    case IMPLIES :
      s1 = get_ctl_left_son(d, f);
      s2 = get_ctl_right_son(d, f);
      ns1 = normalize_ctl_formula(d, s1);
      ns2 = normalize_ctl_formula(d, s2);

      if (s1 == ns1 && s2 == ns2) {
	incref(d, f);
	return f;
      }

      return new_ctl_formula_op(d, get_ctl_type(d, f), ns1, ns2);
    case NOT :
      s1 = get_ctl_son(d, f);
      ns1 = normalize_ctl_formula(d, s1);

      if (ns1 == s1) {
	incref(d, f);
	return f;
      }

      return new_ctl_formula_op(d, NOT, ns1, CTL_NO_FORMULA);
    default :
      /* one can't come here */
      fprintf(stderr, "ctl.c:normalize_ctl_formula (line %d) : error : unreachable code reached !?\n", __LINE__);
      abort();
  }
}

void print_formula(ctl_data *d, variables *v, ctl_formula f)
{
  switch (get_ctl_type(d, f)) {
    case ATOM :
      fprintf(stdout, "%s", get_variable_name(v, get_ctl_variable(d, f)));
      return;
    case ALL_PATH :
      fprintf(stdout, "A(");
      print_formula(d, v, get_ctl_son(d, f));
      fprintf(stdout, ")");
      return;
    case ONE_PATH :
      fprintf(stdout, "E(");
      print_formula(d, v, get_ctl_son(d, f));
      fprintf(stdout, ")");
      return;
    case ALWAYS :
      fprintf(stdout, "G(");
      print_formula(d, v, get_ctl_son(d, f));
      fprintf(stdout, ")");
      return;
    case EVENTUALLY :
      fprintf(stdout, "F(");
      print_formula(d, v, get_ctl_son(d, f));
      fprintf(stdout, ")");
      return;
    case UNTIL :
      fprintf(stdout, "(");
      print_formula(d, v, get_ctl_left_son(d, f));
      fprintf(stdout, ")U(");
      print_formula(d, v, get_ctl_right_son(d, f));
      fprintf(stdout, ")");
      return;
    case NEXT :
      fprintf(stdout, "X(");
      print_formula(d, v, get_ctl_son(d, f));
      fprintf(stdout, ")");
      return;
    case AND :
      fprintf(stdout, "(");
      print_formula(d, v, get_ctl_left_son(d, f));
      fprintf(stdout, ").(");
      print_formula(d, v, get_ctl_right_son(d, f));
      fprintf(stdout, ")");
      return;
    case OR :
      fprintf(stdout, "(");
      print_formula(d, v, get_ctl_left_son(d, f));
      fprintf(stdout, ")+(");
      print_formula(d, v, get_ctl_right_son(d, f));
      fprintf(stdout, ")");
      return;
    case XOR :
      fprintf(stdout, "(");
      print_formula(d, v, get_ctl_left_son(d, f));
      fprintf(stdout, ")^(");
      print_formula(d, v, get_ctl_right_son(d, f));
      fprintf(stdout, ")");
      return;
    case NOT :
      fprintf(stdout, "!(");
      print_formula(d, v, get_ctl_son(d, f));
      fprintf(stdout, ")");
      return;
    case IMPLIES :
      fprintf(stdout, "(");
      print_formula(d, v, get_ctl_left_son(d, f));
      fprintf(stdout, ")->(");
      print_formula(d, v, get_ctl_right_son(d, f));
      fprintf(stdout, ")");
      return;
    case CTL_TRUE :
      fprintf(stdout, "True");
      return;
    case CTL_FALSE :
      fprintf(stdout, "False");
      return;
  }
}
