/*
 * A hash table - more or less generic
 *---
 * to my angel
 * public domain
 * march 2002
 */

#include "hash.h"

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "jenkins_hash.h"
#include "mem.h"
#include "verbose.h"

static inline int bit_on(unsigned char *c, unsigned long i)
{
  register unsigned long ii = i >> 3;
  register int iii = i & 7;

  return c[ii] & (1<<iii);
}

static inline void set_bit(unsigned char *c, unsigned long i)
{
  register unsigned long ii = i >> 3;
  register int iii = i & 7;

  c[ii] |= (1 << iii);
}

#define hash_get_data(h, hash) ((hash_data *)((char *)(h)->data + hash * (h)->elem_size))

/* h->flag must be used to test if all is ok */
void init_hash(hash_table *h, int elem_size, int init_size,
	       hash_compare_function compare, void *compare_pointer,
	       hash_hash_function hash, void *hash_pointer,
	       hash_get_key_function get_key, void *get_key_pointer)
{
  memset(h, 0, sizeof(hash_table));

  h->elem_size = elem_size;
  h->compare = compare;
  h->compare_pointer = compare == HASH_DEFAULT_COMPARE ? (void *)h : compare_pointer;
  h->hash = hash;
  h->hash_pointer = hash == HASH_DEFAULT_HASH ? (void *)h : hash_pointer;
  h->get_key = get_key;
  h->get_key_pointer = get_key == HASH_DEFAULT_GET_TRUE_DATA ? (void *)h : get_key_pointer;

  h->data = getmem(elem_size * init_size);
  h->data_busy = getmem(sizeof(unsigned char) * init_size / 8);
  h->maxsize = init_size;
  memset(h->data_busy, 0, sizeof(unsigned char) * init_size / 8);

  h->flag = HASH_NO_ERROR;
}

void free_hash(hash_table *h)
{
  print("hashstat", stdout, "hash stat : %ld collisions, final hash size %ld, nb items %ld\n", h->hash_collision, h->maxsize, h->size);

  if (h->maxsize) {
    freemem(h->data);
    freemem(h->data_busy);
  }

  memset(h, 0, sizeof(hash_table));

  h->flag = HASH_NO_ERROR;
}

/* when collision, we call hash_next to get the next hash value */
static unsigned long hash_next(hash_table *h, unsigned long hash)
{
  /* some stats */
  h->hash_collision++;

  return hash+h->hash_random;
}

static void rehash(hash_table *h)
{
  unsigned long i;
  unsigned long hash;
  unsigned long index;
  unsigned long oldsize = h->maxsize;
  void *old_data = h->data;
  void *od2 = old_data;
  unsigned char *old_busy = h->data_busy;
  unsigned char *ob2 = old_busy;

  h->maxsize *= 2;
  if (h->maxsize == 0x80000000) {
    fprintf(stderr, "hash.c:rehash(line %d): fatal error, out of memory\n", __LINE__);
    abort();
  }

  h->data = getmem(h->maxsize * h->elem_size);
  h->data_busy = getmem(h->maxsize * sizeof(unsigned char) / 8);

  memset(h->data_busy, 0, sizeof(unsigned char) * h->maxsize / 8);

  for (i=0; i<oldsize; i++, (char *)old_data += h->elem_size) {
    if (!bit_on(old_busy, i))
      continue;

    hash = h->hash(h->hash_pointer, h->get_key(h->get_key_pointer, (hash_data *)old_data));
    index = hash & (h->maxsize - 1);

    /* collision handling */
    while (bit_on(h->data_busy, index)) {
      hash = hash_next(h, hash);
      index = hash & (h->maxsize - 1);
    }

    set_bit(h->data_busy, index);
    memcpy(hash_get_data(h, index), old_data, h->elem_size);
  }

  freemem(od2);
  freemem(ob2);
}

hid hash_get_new_slot(hash_table *h, new_data *d)
{
  unsigned long hash;
  unsigned long index;

  if (h->size * 4 >= h->maxsize * 3)
    rehash(h);

  hash = h->hash(h->hash_pointer, d);
  index = hash & (h->maxsize - 1);

  /* collision detection */
  while (bit_on(h->data_busy, index)) {
    if (h->compare(h->compare_pointer, d, hash_get_data(h, index)) == 1) {
      h->flag = HASH_ALREADY_PRESENT;
      return index;
    }
    hash = hash_next(h, hash);
    index = hash & (h->maxsize - 1);
  }

  set_bit(h->data_busy, index);

  h->size++;

  h->flag = HASH_NO_ERROR;

  return index;
}

hid hash_get_slot(hash_table *h, new_data *d)
{
  unsigned long hash = h->hash(h->hash_pointer, d);
  unsigned long index = hash & (h->maxsize - 1);

  while (bit_on(h->data_busy, index)) {
    if (h->compare(h->compare_pointer, d, hash_get_data(h, index)) == 1) {
      h->flag = HASH_NO_ERROR;
      return index;
    }
    hash = hash_next(h, hash);
    index = hash & (h->maxsize - 1);
  }

  h->flag = HASH_NOT_PRESENT;
  return 0;
}

void hash_feed_slot(hash_table *h, hid id, hash_data *d)
{
  memcpy(hash_get_data(h, id), d, h->elem_size);

  h->flag = HASH_NO_ERROR;
}

void hash_free_slot(hash_table *h, hid id)
{
  fprintf(stderr, "hash.c:hash_free_slot(line %d): error, you can't remove anything from this hash table\n", __LINE__);
  abort();
}

hash_data *hash_get_slot_data(hash_table *h, hid id)
{
  h->flag = HASH_NO_ERROR;

  return hash_get_data(h, id);
}

void hash_start(hash_table *h)
{
  h->hash_current = h->size_current = 0;
  h->hash_random = 0;
}

/* functions for hash calculation */
void hash_update(hash_table *h, void *d, unsigned long s)
{
  /* some "randomness" used for collision handling */
  h->hash_random += *(unsigned char *)d;
  if (s>=2) h->hash_random += *((unsigned char *)d + 1);
  if (s>=3) h->hash_random += *((unsigned char *)d + 2);
  if (s>=4) h->hash_random += *((unsigned char *)d + 3);

  h->hash_current = jenkins_hash((unsigned char *)d, s, h->hash_current);
  h->size_current += s;
}

unsigned long hash_end(hash_table *h)
{
  /* a linear search will be used, but wheigted by hash function and with some entropy from the key (see hash_update) */
  h->hash_random += (h->hash_current<<1)+(h->hash_current<<4)+(h->hash_current<<10);
  h->hash_random |= 1;

  return h->hash_current;
}

unsigned long hash_simple(hash_table *h, void *d, unsigned long s)
{
  hash_start(h);
  hash_update(h, d, s);
  return hash_end(h);
}

/* default functions */

/* return 1 if nd == dd, 0 if not */
int hash_default_compare(hash_table *h, new_data *nd, hash_data *dd)
{
  return memcmp(nd, dd, h->elem_size) == 0;
}

unsigned long hash_default_hash(hash_table *h, new_data *d)
{
  return hash_simple(h, d, h->elem_size);
}

new_data *hash_default_get_key(void *unused, hash_data *d)
{
  return d;
}
