/*
 * cnf2mines, transform a cnf into a minesweeper position
 * june 2002
 * public domain
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#ifdef DO_PNG_DUMP
#include <png.h>
#include <unistd.h>
#endif

#include "trans.h"
#include "ctl.h"
#include "clauses.h"
#include "bmc.h"
#include "verbose.h"
#include "mem.h"

/*
 * a cnf parser (several functions)
 * reads in DIMACS file format (only 'p cnf' ones)
 */

transition_tree read_clause(FILE *f, trans_data *d)
{
  int nb=0;
  transition_tree ret = TRANS_UNDEF, t;

  while (!feof(f)) {
    if (fscanf(f, "%d", &nb) != 1) {
      fprintf(stderr, "error, number wanted\n");
      return TRANS_UNDEF;
    }

    if (nb == 0)
      break;

    if (nb<0) {
      t = new_transition_tree_var(d, -nb);
      t = new_transition_tree_op(d, NOT, t, TRANS_UNDEF);
    } else
      t = new_transition_tree_var(d, nb);

    if (ret == TRANS_UNDEF)
      ret = t;
    else
      ret = new_transition_tree_op(d, OR, ret, t);
  }

  return ret;
}

#define ERROR	-1
#define COMMENT	1
#define TYPE	2
#define CLAUSE	3
#define END	4

int type_line(FILE *f)
{
  int c;

  do c = fgetc(f); while (isspace(c) && c != EOF);

  if (c == 'c' || c == 'C')
    return COMMENT;

  if (c == 'p' || c == 'P')
    return TYPE;

  if (c == '-' || isdigit(c)) {
    ungetc(c, f);
    return CLAUSE;
  }

  if (c == EOF)
    return END;

  fprintf(stderr, "wrong line type\n");
  return ERROR;
}

void erase_line(FILE *f)
{
  int c;

  do c = fgetc(f); while (c != EOF && c != '\n');
}

int vars_line(FILE *f)
{
  int c, v;

  do c = fgetc(f); while (isspace(c) && c != EOF);

  /* 'cnf' must start the line */
  if (c != 'c') return -1;
  c = fgetc(f); if (c != 'n') return -1;
  c = fgetc(f); if (c != 'f') return -1;
  c = fgetc(f); if (!isspace(c)) return -1;

  if (fscanf(f, "%d", &v) != 1) {
    fprintf(stderr, "wrong 'p cnf x x' line\n");
    return -1;
  }

  if (fscanf(f, "%d", &c) != 1) {
    fprintf(stderr, "wrong 'p cnf x x' line\n");
    return -1;
  }

  return v;
}

transition_tree read_cnf(FILE *f, trans_data *d)
{
  int nbvars;
  int c;
  transition_tree t = TRANS_UNDEF;

  /* remove comments lines */
  while (!feof(f)) {
    if ((c=type_line(f)) != COMMENT)
      break;
    erase_line(f);
  }

  if (feof(f))
    return t;

  if (c != TYPE) {
    fprintf(stderr, "wrong line, should be TYPE\n");
    return t;
  }

  nbvars = vars_line(f);
  if (nbvars <= 0) {
    fprintf(stderr, "we must have at least one var\n");
    return t;
  }

  while (!feof(f)) {
    transition_tree t2;

    int c;
    do c = fgetc(f); while (c != EOF && isspace(c));
    if (c == EOF) return t;
    ungetc(c, f);

    t2 = read_clause(f, d);

    if (t2 == TRANS_UNDEF)
      return t2;

    if (t == TRANS_UNDEF)
      t = t2;
    else
      t = new_transition_tree_op(d, AND, t, t2);
  }

  return t;
}

/*
 * the transformation from cnf to minesweeper
 * several functions
 */

void get_size(trans_data *d, unsigned long *width, unsigned long *height)
{
  unsigned long w=4, h=0;
  unsigned long curcol = 2;
  unsigned long i;
  _transition_tree *t = d->trans_tree;

  for (i=0; i<d->size; i++, t++) {
    switch(t->op) {
      case VAR :
        /* new wire */
        t->x = 0;
        t->y = h;
        t->wire_line = h+1;
        h += 7;
        break;

      case NOT :
        t->x = curcol;
        t->y = h;
        t->wire_line = h+3;
        h += 9;
        w += 7;
        curcol += 7;
        break;

      case AND :
        t->x = curcol;
        t->y = h;
        t->wire_line = h+4;
        h += 10;
        w += 14;
        curcol += 14;
        break;

      case OR :
        t->x = curcol;
        t->y = h;
        t->wire_line = h+3;
        h += 9;
        w += 10;
        curcol += 10;
        break;

      default : fprintf(stderr, "error type\n"); abort();
    }
  }

  *width = w;
  *height = h;
}

typedef struct {
  char type;
  char x, y;
} base;

void wire(base *p)
{
  if (p->type) abort();

  p->type = '>';
}

void vwire(base *p)
{
  if (p->type) abort();

  p->type = 'v';
}

void split(base *p, int w)
{
  int i;

  p--;
  p -= w;

  for (i=0; i<5; i++, p+=w) {
    if (p->type != (i==1 ? '>' : 0)) abort();
    p->type = 's';
    p->x = 0;
    p->y = i;

    if ((p+1)->type != (i==1 ? '>' : 0)) abort();
    (p+1)->type = 's';
    (p+1)->x = 1;
    (p+1)->y = i;

    if ((p+2)->type != (i==1 ? '>' : 0)) abort();
    (p+2)->type = 's';
    (p+2)->x = 2;
    (p+2)->y = i;
  }
}

void cross(base *p, int w)
{
  int i;

  p -= 2;
  p -= w;

  for (i=0; i<7; i++, p+=w) {
    if (p->type != (i==1 ? '>' : 0)) abort();
    p->type = 'c';
    p->x = 0;
    p->y = i;

    if ((p+1)->type != (i==1 ? '>' : 0)) abort();
    (p+1)->type = 'c';
    (p+1)->x = 1;
    (p+1)->y = i;

    if ((p+2)->type != (i==0 ? 'v' : (i==1 ? '>' : 0))) abort();
    (p+2)->type = 'c';
    (p+2)->x = 2;
    (p+2)->y = i;

    if ((p+3)->type != (i==1 ? '>' : 0)) abort();
    (p+3)->type = 'c';
    (p+3)->x = 3;
    (p+3)->y = i;

    if ((p+4)->type != (i==1 ? '>' : 0)) abort();
    (p+4)->type = 'c';
    (p+4)->x = 4;
    (p+4)->y = i;
  }
}

void or(base *p, int w)
{
  int i;

  for (i=0; i<4; i++, p+=w) {
    if (p->type) abort();
    p->type = 'o';
    p->x = 0;
    p->y = i;

    if ((p+1)->type) abort();
    (p+1)->type = 'o';
    (p+1)->x = 1;
    (p+1)->y = i;

    if ((p+2)->type) abort();
    (p+2)->type = 'o';
    (p+2)->x = 2;
    (p+2)->y = i;

    if ((p+3)->type) abort();
    (p+3)->type = 'o';
    (p+3)->x = 3;
    (p+3)->y = i;
  }
}

void not(base *p, int w)
{
  int i;

  for (i=0; i<3; i++, p+=w) {
    if (p->type) abort();
    p->type = 'n';
    p->x = 0;
    p->y = i;

    if ((p+1)->type) abort();
    (p+1)->type = 'n';
    (p+1)->x = 1;
    (p+1)->y = i;

    if ((p+2)->type) abort();
    (p+2)->type = 'n';
    (p+2)->x = 2;
    (p+2)->y = i;
  }
}

void Not(base *p, int w)
{
  int i;

  for (i=0; i<3; i++, p+=w) {
    if (p->type) abort();
    p->type = 'N';
    p->x = 0;
    p->y = i;

    if ((p+1)->type) abort();
    (p+1)->type = 'N';
    (p+1)->x = 1;
    (p+1)->y = i;

    if ((p+2)->type) abort();
    (p+2)->type = 'N';
    (p+2)->x = 2;
    (p+2)->y = i;
  }
}

void turn(base *p, int w)
{
  int i;

  for (i=0; i<5; i++, p+=w) {
    if (p->type) abort();
    p->type = 'L';
    p->x = 0;
    p->y = i;

    if ((p+1)->type) abort();
    (p+1)->type = 'L';
    (p+1)->x = 1;
    (p+1)->y = i;

    if ((p+2)->type) abort();
    (p+2)->type = 'L';
    (p+2)->x = 2;
    (p+2)->y = i;
  }
}

void terminal(base *p, int w)
{
  int i;

  for (i=0; i<3; i++, p+=w) {
    if (p->type) abort();
    p->type = 't';
    p->x = 0;
    p->y = i;

    if ((p+1)->type) abort();
    (p+1)->type = 't';
    (p+1)->x = 1;
    (p+1)->y = i;
  }
}

void Terminal(base *p, int w)
{
  int i;

  for (i=0; i<3; i++, p+=w) {
    if (p->type) abort();
    p->type = 'T';
    p->x = 0;
    p->y = i;

    if ((p+1)->type) abort();
    (p+1)->type = 'T';
    (p+1)->x = 1;
    (p+1)->y = i;
  }
}

void vertical_wire(base *tab, int w, int h, int x, int y1, int y2)
{
  int i;
  base *p = tab;

  p += y1 * w;
  p += x;

  if (y2 < y1) return;

  if (p->type != '>') {
    fprintf(stderr, "wire wanted %d %c\n", p->type, p->type);
    abort();
  }

  split(p, w);

  p += 4 * w;

  for (i=y1+4; i<=y2; i++, p+=w) {
    switch(p->type) {
      case 0 :
	vwire(p);
	break;

      case '>' :
	cross(p, w);
	p += 5 * w;
	i += 5;
	break;

      default :
	fprintf(stderr, "vertical wire inserted at bad place\n");
	abort();
    }
  }
}

void horizontal_wire(base *p, int len)
{
  int i;

  for (i=0; i<len; i++, p++)
    wire(p);
}

void position_var(_transition_tree *t, base *tab, int w, int h)
{
  int i;
  base *p = tab;

  p += t->y * w;

  terminal(p, w);

  p+=2;
  for (i=0; i<w-4; i++, p++)
    wire(p+w);

  Terminal(p, w);
}

void position_or(trans_data *d, _transition_tree *t, base *tab, int w, int h)
{
  base *p = tab;
  int y1, y2;

  p += t->y * w;
  p += t->x;

  turn(p+1, w);
  or(p+2*w+6, w);
  vwire(p+7);
  vwire(p+7+w);
  wire(p+3*w+4);
  wire(p+3*w+5);

  horizontal_wire(tab + t->y * w + t->x + 3*w + 10, w-2-10-t->x);

  Terminal(tab + t->y * w + w * 3 - 2, w);

  y1 = d->trans_tree[t->val.op_two.l].wire_line;
  y2 = d->trans_tree[t->val.op_two.r].wire_line;

  vertical_wire(tab, w, h, t->x+2, y1, t->y-1);
  vertical_wire(tab, w, h, t->x+7, y2, t->y-1);
}

void position_and(trans_data *d, _transition_tree *t, base *tab, int w, int h)
{
  base *p = tab;
  int y1, y2;

  p += t->y * w;
  p += t->x;

  turn(p+1+w, w);
  not(p+3*w+4, w);
  or(p+3*w+7, w);
  not(p+3*w+11, w);
  Not(p+7, w);
  vwire(p+2);

  horizontal_wire(tab + t->y * w + t->x + 4*w + 14, w-2-14-t->x);

  Terminal(tab + t->y * w + w * 4 - 2, w);

  y1 = d->trans_tree[t->val.op_two.l].wire_line;
  y2 = d->trans_tree[t->val.op_two.r].wire_line;

  vertical_wire(tab, w, h, t->x+2, y1, t->y-1);
  vertical_wire(tab, w, h, t->x+8, y2, t->y-1);
}

void position_not(trans_data *d, _transition_tree *t, base *tab, int w, int h)
{
  base *p = tab;
  int y1;

  p += t->y * w;
  p += t->x;

  turn(p+1, w);
  not(p+2*w+4, w);

  horizontal_wire(tab + t->y * w + t->x + 3*w + 7, w-2-7-t->x);

  Terminal(tab + t->y * w + w * 3 - 2, w);

  y1 = d->trans_tree[t->val.op_one.s].wire_line;

  vertical_wire(tab, w, h, t->x+2, y1, t->y-1);
}

void clear_last_and(trans_data *d, _transition_tree *t, base *tab, int w, int h)
{
  int i;
  base *p = tab + t->y * w + w * 4 - 2;

  for (i=0; i<3; i++, p+=w) {
    p->type = 0;
    (p+1)->type = 0;
  }
}

void position(trans_data *d, base *tab, int w, int h)
{
  int i;
  _transition_tree *t = d->trans_tree;

  for (i=0; i<d->size; i++, t++) {
    switch(t->op) {
      case VAR :
	position_var(t, tab, w, h);
	break;

      case NOT :
        position_not(d, t, tab, w, h);
	break;

      case AND :
        position_and(d, t, tab, w, h);
	break;

      case OR :
        position_or(d, t, tab, w, h);
	break;

      default : fprintf(stderr, "unhandled type %d\n", t->op); break;
    }
  }

  t--;

  if (t->op != AND) {
    fprintf(stderr, "fatal error\n");
    abort();
  }

  clear_last_and(d, t, tab, w, h);
}

/*
 * printing function and needed data
 */

char d_wire[9] = {
'1', '1', '1',
'x', 'x', '1',
'1', '1', '1'
};

char d_vwire[9] = {
'1', 'x', '1',
'1', 'x', '1',
'1', '1', '1'
};

/* [y][x] */
char d_not[3][3][9] = {
{
{
' ', ' ', ' ',
' ', ' ', ' ',
' ', ' ', '1'
},
{
' ', ' ', ' ',
' ', ' ', ' ',
'1', '2', '1'
},
{
' ', ' ', ' ',
' ', ' ', ' ',
'2', '1', '1'
}
},

{
{
'1', '1', '2',
'x', 'x', '3',
'1', '1', '2'
},
{
'*', '3', '*',
'x', '5', 'x',
'*', '3', '*'
},
{
'3', '*', '2',
'5', 'x', '3',
'3', '*', '2'
}
},

{
{
' ', ' ', '1',
' ', ' ', ' ',
' ', ' ', ' '
},
{
'1', '2', '1',
' ', ' ', ' ',
' ', ' ', ' '
},
{
'2', '1', '1',
' ', ' ', ' ',
' ', ' ', ' '
}
}
};

char d_Not[3][3][9] = {
{
{
' ', ' ', ' ',
' ', ' ', ' ',
' ', ' ', '1'
},
{
'1', 'x', '1',
'1', 'x', '1',
'2', '3', '2'
},
{
' ', ' ', ' ',
' ', ' ', ' ',
'1', ' ', ' '
}
},

{
{
' ', ' ', '1',
' ', ' ', '2',
' ', ' ', '1'
},
{
'*', 'x', '*',
'3', '5', '3',
'*', 'x', '*'
},
{
'1', ' ', ' ',
'2', ' ', ' ',
'1', ' ', ' '
}
},

{
{
' ', ' ', '2',
' ', ' ', '1',
' ', ' ', '1'
},
{
'3', '5', '3',
'*', 'x', '*',
'2', '3', '2'
},
{
'2', ' ', ' ',
'1', ' ', ' ',
'1', ' ', ' '
}
}
};

char d_terminal[3][2][9] = {
{
{
' ', ' ', ' ',
' ', ' ', ' ',
' ', '1', '1'
},
{
' ', ' ', ' ',
' ', ' ', ' ',
'1', ' ', ' '
}
},

{
{
' ', '2', '*',
' ', '3', '*',
' ', '2', '*'
},
{
'3', '1', '1',
'x', 'x', '1',
'3', '1', '1'
}
},

{
{
' ', '1', '1',
' ', ' ', ' ',
' ', ' ', ' '
},
{
'1', ' ', ' ',
' ', ' ', ' ',
' ', ' ', ' '
}
}
};

char d_Terminal[3][2][9] = {
{
{
' ', ' ', ' ',
' ', ' ', ' ',
' ', '1', '1'
},
{
' ', ' ', ' ',
' ', ' ', ' ',
'1', ' ', ' '
}
},

{
{
'1', '3', '*',
'x', 'x', '*',
'1', '3', '*'
},
{
'2', ' ', ' ',
'3', ' ', ' ',
'2', ' ', ' '
}
},

{
{
' ', '1', '1',
' ', ' ', ' ',
' ', ' ', ' '
},
{
'1', ' ', ' ',
' ', ' ', ' ',
' ', ' ', ' '
}
}
};

char d_turn[5][3][9] =
{
/* 1 */
{
{
' ', ' ', ' ',
' ', ' ', ' ',
' ', '1', '1'
},
{
'1', 'x', '1',
'1', 'x', '1',
'2', '2', '2'
},
{
' ', ' ', ' ',
' ', ' ', ' ',
'1', ' ', ' '
}
},

/* 2 */
{
{
' ', '2', '*',
' ', '2', '*',
' ', '1', '2'
},
{
'3', 'x', '*',
'x', '3', '*',
'2', '2', '1'
},
{
'2', ' ', ' ',
'2', ' ', ' ',
'1', ' ', ' '
}
},

/* 3 */
{
{
' ', ' ', '1',
' ', ' ', '1',
' ', ' ', '2'
},
{
'x', '1', ' ',
'x', '1', ' ',
'2', '1', ' '
},
{
' ', ' ', ' ',
' ', ' ', ' ',
' ', ' ', ' '
}
},

/* 4 */
{
{
' ', '2', '*',
' ', '2', '*',
' ', '1', '3'
},
{
'x', '2', '1',
'*', 'x', '2',
'*', '*', '2'
},
{
'1', '1', '1',
'x', 'x', '1',
'1', '1', '1'
}
},

/* 5 */
{
{
' ', ' ', '1',
' ', ' ', ' ',
' ', ' ', ' '
},
{
'2', '2', '1',
' ', ' ', ' ',
' ', ' ', ' '
},
{
' ', ' ', ' ',
' ', ' ', ' ',
' ', ' ', ' '
}
}
};

char d_split[5][3][9] = {
/* 1 */
{
{
' ', ' ', ' ',
' ', ' ', '1',
' ', ' ', '1'
},
{
' ', ' ', ' ',
'2', '3', '2',
'*', '*', '*'
},
{
' ', ' ', ' ',
'1', ' ', ' ',
'2', '1', '1'
}
},

/* 2 */
{
{
'1', '1', '2',
'x', 'x', '1',
'1', '1', '1'
},
{
'3', 'x', '3',
'x', '2', 'x',
'1', 'x', '1'
},
{
'3', '*', '2',
'3', 'x', '3',
'2', '*', '2'
}
},

/* 3 */
{
{
' ', ' ', '1',
' ', ' ', '1',
' ', ' ', '2'
},
{
'2', '3', '2',
'*', 'x', '*',
'3', '5', '3'
},
{
'2', '1', '1',
'1', ' ', ' ',
'2', ' ', ' '
}
},

/* 4 */
{
{
' ', ' ', '1',
' ', ' ', '2',
' ', ' ', '1'
},
{
'*', 'x', '*',
'3', '5', '3',
'*', 'x', '*'
},
{
'1', ' ', ' ',
'2', ' ', ' ',
'1', ' ', ' '
}
},

/* 5 */
{
{
' ', ' ', '2',
' ', ' ', '1',
' ', ' ', '1'
},
{
'3', '5', '3',
'*', 'x', '*',
'2', '3', '2'
},
{
'2', ' ', ' ',
'1', ' ', ' ',
'1', ' ', ' '
}
}
};

char d_cross[7][5][9] = {
/* 1 */
{
{
' ', ' ', ' ',
' ', '1', '2',
' ', '1', '*'
},
{
' ', ' ', ' ',
'3', '2', '1',
'*', '*', '2'
},
{
'1', 'x', '1',
'1', 'x', '1',
'2', '3', '2'
},
{
' ', ' ', ' ',
' ', '1', '2',
'1', '1', '*'
},
{
' ', ' ', ' ',
'3', '2', '1',
'*', '*', '1'
}
},

/* 2 */
{
{
'1', '3', '5',
'x', 'x', '*',
'1', '3', '4'
},
{
'x', '5', '4',
'*', '*', 'x',
'x', '*', '*'
},
{
'*', 'x', '*',
'x', '6', 'x',
'*', 'x', '*'
},
{
'2', '2', '3',
'x', '1', 'x',
'2', '1', '1'
},
{
'x', '3', '2',
'2', 'x', '1',
'x', '1', '1'
}
},

/* 3 */
{
{
' ', '1', '*',
' ', '1', '2',
' ', ' ', '1'
},
{
'4', 'x', '*',
'*', '5', '*',
'3', '*', 'x'
},
{
'5', '3', '2',
'4', 'x', '2',
'x', 'x', 'x'
},
{
'1', ' ', '1',
'1', '1', '1',
'x', '2', 'x'
},
{
'2', '2', '1',
'x', '*', '2',
'5', '*', '2'
}
},

/* 4 */
{
{
' ', ' ', ' ',
' ', ' ', ' ',
' ', ' ', ' '
},
{
'2', '*', '3',
'1', '1', '1',
' ', ' ', ' '
},
{
'2', '2', '2',
'1', 'x', '1',
'1', 'x', '1'
},
{
'1', '2', '*',
' ', '1', '2',
' ', ' ', ' '
},
{
'*', '2', '1',
'2', '1', ' ',
' ', ' ', ' '
}
},

/* 5 */
{
{
' ', ' ', ' ',
' ', ' ', ' ',
' ', ' ', ' '
},
{
' ', ' ', '1',
' ', ' ', '1',
' ', ' ', '2'
},
{
'2', '3', '2',
'*', 'x', '*',
'3', '5', '3'
},
{
'1', ' ', ' ',
'1', ' ', ' ',
'2', ' ', ' '
},
{
' ', ' ', ' ',
' ', ' ', ' ',
' ', ' ', ' '
}
},

/* 6 */
{
{
' ', ' ', ' ',
' ', ' ', ' ',
' ', ' ', ' '
},
{
' ', ' ', '1',
' ', ' ', '2',
' ', ' ', '1'
},
{
'*', 'x', '*',
'3', '5', '3',
'*', 'x', '*'
},
{
'1', ' ', ' ',
'2', ' ', ' ',
'1', ' ', ' '
},
{
' ', ' ', ' ',
' ', ' ', ' ',
' ', ' ', ' '
}
},

/* 7 */
{
{
' ', ' ', ' ',
' ', ' ', ' ',
' ', ' ', ' '
},
{
' ', ' ', '2',
' ', ' ', '1',
' ', ' ', '1'
},
{
'3', '5', '3',
'*', 'x', '*',
'2', '3', '2'
},
{
'2', ' ', ' ',
'1', ' ', ' ',
'1', ' ', ' '
},
{
' ', ' ', ' ',
' ', ' ', ' ',
' ', ' ', ' '
}
}
};

char d_or[4][4][9] = {
/* 1 */
{
{
' ', ' ', ' ',
' ', ' ', ' ',
' ', ' ', '1'
},
{
'1', 'x', '1',
'1', 'x', '1',
'2', '3', '2'
},
{
' ', ' ', ' ',
' ', '1', '2',
'1', '1', '*'
},
{
' ', ' ', ' ',
'3', '2', '1',
'*', '*', '1'
}
},

/* 2 */
{
{
'1', '1', '2',
'x', 'x', '3',
'1', '1', '2'
},
{
'*', 'x', '*',
'x', '6', 'x',
'*', 'x', '*'
},
{
'2', '2', '3',
'x', '1', 'x',
'5', '4', '3'
},
{
'x', '3', '2',
'2', 'x', '1',
'x', '2', '2'
}
},

/* 3 */
{
{
' ', ' ', '2',
' ', ' ', '2',
' ', ' ', '2'
},
{
'4', '*', '*',
'*', 'x', 'x',
'*', '*', '3'
},
{
'*', '*', '*',
'x', 'x', 'x',
'2', '3', '*'
},
{
'4', '*', '1',
'*', '3', '1',
'*', '2', ' '
}
},

/* 4 */
{
{
' ', ' ', '1',
' ', ' ', ' ',
' ', ' ', ' '
},
{
'2', '2', '1',
' ', ' ', ' ',
' ', ' ', ' '
},
{
' ', '1', '2',
' ', ' ', ' ',
' ', ' ', ' '
},
{
'2', '1', ' ',
' ', ' ', ' ',
' ', ' ', ' '
}
}
};

#ifdef DO_PNG_DUMP
/* some data for png dumping */
#include "images.h"

int colors[11][3] = {
{132, 130, 132},
{198, 195, 198},
{0, 0, 255},
{0, 130, 0},
{255, 0, 0},
{0, 0, 132},
{132, 0, 0},
{0, 130, 132},
{132, 0, 132},
{0, 0, 0},
{255, 255, 255}
};

void put_char(char *row, int pos, char c, int l)
{
  char *from;

  switch(c) {
    case ' ' : from = _0; break;
    case '1' : from = _1; break;
    case '2' : from = _2; break;
    case '3' : from = _3; break;
    case '4' : from = _4; break;
    case '5' : from = _5; break;
    case '6' : from = _6; break;
    case '7' : from = _7; break;
    case '8' : from = _8; break;
    case '*' : from = _marked; break;
    case 'x' : from = _unknown; break;
    default : abort();
  }
  from += l*16;
  memcpy(row+pos, from, 16);
}

void dump_to_png(base *tab, int w, int h, char *png_file)
{
  FILE *f;
  png_structp s;
  png_infop t;
  png_color pal[11];
  char *row = getmem(w*3*16);
  int rowp;
  int i, k, j, l;
  int x, y;

  f = fopen(png_file, "wb");
  if (!f) {
    perror(png_file);
    return;
  }

  s = png_create_write_struct(PNG_LIBPNG_VER_STRING, 0, 0, 0);
  if (!s)
    goto error;

  t = png_create_info_struct(s);
  if (!t)
    goto error2;

  png_init_io(s, f);

  png_set_compression_level(s, Z_BEST_COMPRESSION);

  png_set_IHDR(s, t,
                /*width*/w*3*16,
                /*height*/h*3*16,
                /*bit depth*/8,
                /*color_type*/PNG_COLOR_TYPE_PALETTE,
                PNG_INTERLACE_NONE,
                PNG_COMPRESSION_TYPE_DEFAULT,
                PNG_FILTER_TYPE_DEFAULT);

  for (i=0; i<11; i++) {
    pal[i].red = colors[i][0];
    pal[i].green = colors[i][1];
    pal[i].blue = colors[i][2];
  }
  png_set_PLTE(s, t, pal, 11);

  png_write_info(s, t);

  for (i=0; i<h; i++, tab+=w) {
    for (k=0; k<3; k++)
    for (l=0; l<16; l++) {
      rowp = 0;
      for (j=0; j<w; j++) {
	x = (tab+j)->x;
	y = (tab+j)->y;
        switch ((tab+j)->type) {
	  case 't' :
	    put_char(row, rowp, d_terminal[y][x][k*3+0], l);
	    rowp+=16;
	    put_char(row, rowp, d_terminal[y][x][k*3+1], l);
	    rowp+=16;
	    put_char(row, rowp, d_terminal[y][x][k*3+2], l);
	    rowp+=16;
	    break;
	  case 'T' :
	    put_char(row, rowp, d_Terminal[y][x][k*3+0], l);
	    rowp+=16;
	    put_char(row, rowp, d_Terminal[y][x][k*3+1], l);
	    rowp+=16;
	    put_char(row, rowp, d_Terminal[y][x][k*3+2], l);
	    rowp+=16;
	    break;
	  case 'n' :
	    put_char(row, rowp, d_not[y][x][k*3+0], l);
	    rowp+=16;
	    put_char(row, rowp, d_not[y][x][k*3+1], l);
	    rowp+=16;
	    put_char(row, rowp, d_not[y][x][k*3+2], l);
	    rowp+=16;
	    break;
	  case 'N' :
	    put_char(row, rowp, d_Not[y][x][k*3+0], l);
	    rowp+=16;
	    put_char(row, rowp, d_Not[y][x][k*3+1], l);
	    rowp+=16;
	    put_char(row, rowp, d_Not[y][x][k*3+2], l);
	    rowp+=16;
	    break;
	  case '>' :
	    put_char(row, rowp, d_wire[k*3+0], l);
	    rowp+=16;
	    put_char(row, rowp, d_wire[k*3+1], l);
	    rowp+=16;
	    put_char(row, rowp, d_wire[k*3+2], l);
	    rowp+=16;
	    break;
	  case 'v' :
	    put_char(row, rowp, d_vwire[k*3+0], l);
	    rowp+=16;
	    put_char(row, rowp, d_vwire[k*3+1], l);
	    rowp+=16;
	    put_char(row, rowp, d_vwire[k*3+2], l);
	    rowp+=16;
	    break;
	  case 'c' :
	    put_char(row, rowp, d_cross[y][x][k*3+0], l);
	    rowp+=16;
	    put_char(row, rowp, d_cross[y][x][k*3+1], l);
	    rowp+=16;
	    put_char(row, rowp, d_cross[y][x][k*3+2], l);
	    rowp+=16;
	    break;
	  case 's' :
	    put_char(row, rowp, d_split[y][x][k*3+0], l);
	    rowp+=16;
	    put_char(row, rowp, d_split[y][x][k*3+1], l);
	    rowp+=16;
	    put_char(row, rowp, d_split[y][x][k*3+2], l);
	    rowp+=16;
	    break;
	  case 'L' :
	    put_char(row, rowp, d_turn[y][x][k*3+0], l);
	    rowp+=16;
	    put_char(row, rowp, d_turn[y][x][k*3+1], l);
	    rowp+=16;
	    put_char(row, rowp, d_turn[y][x][k*3+2], l);
	    rowp+=16;
	    break;
	  case 'o' :
	    put_char(row, rowp, d_or[y][x][k*3+0], l);
	    rowp+=16;
	    put_char(row, rowp, d_or[y][x][k*3+1], l);
	    rowp+=16;
	    put_char(row, rowp, d_or[y][x][k*3+2], l);
	    rowp+=16;
	    break;
	  case 0 :
	    put_char(row, rowp, ' ', l);
	    rowp+=16;
	    put_char(row, rowp, ' ', l);
	    rowp+=16;
	    put_char(row, rowp, ' ', l);
	    rowp+=16;
	    break;
	  default :
	    fprintf(stderr, "unhandled type %d %c\n", (tab+j)->type, (tab+j)->type);
	    abort();
	}
      }
      png_write_row(s, row);
    }
  }

  png_write_end(s, t);

  png_destroy_write_struct(&s, &t);

  fclose(f);

  freemem(row);

  return;

error2:
  png_destroy_write_struct(&s, 0);

error:
  freemem(row);
  fclose(f);
  unlink(png_file);
  fprintf(stderr, "an error occured while dumping to png\n");
}
#endif /* DO_PNG_DUMP */

void print_symbolic(FILE *f, base *tab, int w, int h)
{
  int i, j;

  for (i=0; i<h; i++) {
    for (j=0; j<w; j++, tab++)
      fprintf(f, "%c", tab->type ? tab->type : ' ');
    fprintf(f, "\n");
  }
}

void print_tab(FILE *f, base *tab, int w, int h)
{
  int i, k, j;
  int x, y;

  for (i=0; i<h; i++, tab+=w) {
    for (k=0; k<3; k++) {
      for (j=0; j<w; j++) {
	x = (tab+j)->x;
	y = (tab+j)->y;
        switch ((tab+j)->type) {
	  case 't' :
	    fprintf(f, "%c%c%c", d_terminal[y][x][k*3+0], d_terminal[y][x][k*3+1], d_terminal[y][x][k*3+2]);
	    break;
	  case 'T' :
	    fprintf(f, "%c%c%c", d_Terminal[y][x][k*3+0], d_Terminal[y][x][k*3+1], d_Terminal[y][x][k*3+2]);
	    break;
	  case 'n' :
	    fprintf(f, "%c%c%c", d_not[y][x][k*3+0], d_not[y][x][k*3+1], d_not[y][x][k*3+2]);
	    break;
	  case 'N' :
	    fprintf(f, "%c%c%c", d_Not[y][x][k*3+0], d_Not[y][x][k*3+1], d_Not[y][x][k*3+2]);
	    break;
	  case '>' :
	    fprintf(f, "%c%c%c", d_wire[k*3+0], d_wire[k*3+1], d_wire[k*3+2]);
	    break;
	  case 'v' :
	    fprintf(f, "%c%c%c", d_vwire[k*3+0], d_vwire[k*3+1], d_vwire[k*3+2]);
	    break;
	  case 'c' :
	    fprintf(f, "%c%c%c", d_cross[y][x][k*3+0], d_cross[y][x][k*3+1], d_cross[y][x][k*3+2]);
	    break;
	  case 's' :
	    fprintf(f, "%c%c%c", d_split[y][x][k*3+0], d_split[y][x][k*3+1], d_split[y][x][k*3+2]);
	    break;
	  case 'L' :
	    fprintf(f, "%c%c%c", d_turn[y][x][k*3+0], d_turn[y][x][k*3+1], d_turn[y][x][k*3+2]);
	    break;
	  case 'o' :
	    fprintf(f, "%c%c%c", d_or[y][x][k*3+0], d_or[y][x][k*3+1], d_or[y][x][k*3+2]);
	    break;
	  case 0 : fprintf(f, "   "); break;
	  default :
	    fprintf(stderr, "unhandled type %d %c\n", (tab+j)->type, (tab+j)->type);
	    abort();
	}
      }
      fprintf(f, "\n");
    }
  }
}

/* useless, but the prog wants it to compile (because all this is a big rip) */
checker the_checker;

#define WE_DO_PNG	1
#define WE_DO_TXT	2
#define WE_DO_SYMBOLIC	4

void fatal_error(char *f)
{
  fprintf(stderr, "usage : %s [input_file] [options]\n", f);
  fprintf(stderr, "    options are :\n");
  fprintf(stderr, "        -t <file> : to output as a text\n");
  fprintf(stderr, "        -s <file> : to output symbolically\n");
#ifdef DO_PNG_DUMP
  fprintf(stderr, "        -p <file> : to output into a png file\n");
#endif
  fprintf(stderr, "you can use 2 or 3 of those options at the same time\n");
  fprintf(stderr, "if you use only -t or -s, you may omit <file>, stdout will be used\n");
  fprintf(stderr, "if no input file given, stdin will be used\n");
  exit(0);
}

int main(int argc, char **argv)
{
  int i;
  trans_data d;
  transition_tree t;
  verbose v;
  unsigned long w, h;
  unsigned long size;
  base *tab;
  int we_do = 0;
  char *file_txt=0, *file_symbolic=0;
#ifdef DO_PNG_DUMP
  char *file_png=0;
#endif
  FILE *f;
  char *in_file = 0;

  if (argc <= 1)
    fatal_error(argv[0]);

  for (i=1; i<argc; i++) {
#ifdef DO_PNG_DUMP
    if (!strcmp(argv[i], "-p")) {
      i++;
      if (i >= argc) fatal_error(argv[0]);
      file_png = argv[i];
      we_do |= WE_DO_PNG;
      continue;
    }
#endif /* DO_PNG_DUMP */
    if (!strcmp(argv[i], "-t")) {
      i++;
      if (i >= argc) {
	/* if no file name given, and if '-t' is the only arg, then dump to stdout */
	if (we_do) fatal_error(argv[0]);
        we_do |= WE_DO_TXT;
	continue;
      }
      file_txt = argv[i];
      we_do |= WE_DO_TXT;
      continue;
    }
    if (!strcmp(argv[i], "-s")) {
      i++;
      if (i >= argc) {
	/* if no file name given, and if '-t' is the only arg, then dump to stdout */
	if (we_do) fatal_error(argv[0]);
        we_do |= WE_DO_SYMBOLIC;
	continue;
      }
      file_symbolic = argv[i];
      we_do |= WE_DO_SYMBOLIC;
      continue;
    }
    /* no option, must be in_file */
    if (in_file) fatal_error(argv[0]);
    in_file = argv[i];
  }

  if (we_do == 0) fatal_error(argv[0]);

  init_verbose(&v);
  //verbose_register(&v, "all");
  init_trans(&d);

  if (!in_file)
    t = read_cnf(stdin, &d);
  else {
    FILE *z = fopen(in_file, "r");
    if (!z) {
      perror(in_file);
      goto end;
    }
    t = read_cnf(z, &d);
    fclose(z);
  }

  get_size(&d, &w, &h);
  fprintf(stderr, "width = %ld, height = %ld\n", w, h);

  size = w * h;
  if (size < w || size < h || size > 1000000000) {
    fprintf(stderr, "problem size too big to generate a minesweeper game, byebye\n");
    goto end;
  }

  tab = getmem(sizeof(base) * size);
  memset(tab, 0, size * sizeof(base));

  position(&d, tab, w, h);

  if (we_do & WE_DO_TXT) {
    f = stdout;
    if (file_txt) {
      f = fopen(file_txt, "w");
      if (!f) {
	perror(file_txt);
	goto after_txt;
      }
    }

    print_tab(f, tab, w, h);

    if (f != stdout) fclose(f);
  }

after_txt:
  if (we_do & WE_DO_SYMBOLIC) {
    f = stdout;
    if (file_symbolic) {
      f = fopen(file_symbolic, "w");
      if (!f) {
	perror(file_symbolic);
	goto after_symbolic;
      }
    }

    print_symbolic(f, tab, w, h);

    if (f != stdout) fclose(f);
  }

after_symbolic:
#ifdef DO_PNG_DUMP
  if (we_do & WE_DO_PNG)
    dump_to_png(tab, w, h, file_png);
#endif

  freemem(tab);

end:
  free_trans(&d);
  free_verbose(&v);

  return 0;
}
