/*
 * factorize - a factorize cnf generator
 * june 2002
 * public domain
 */

#include <stdio.h>
#include <string.h>

#include "trans.h"
#include "ctl.h"
#include "clauses.h"
#include "bmc.h"
#include "verbose.h"
#include "mem.h"

static int curvar=1;
void get_variables(trans_data *d, int *t, int s)
{
  int i;

  for (i=0; i<s; i++, curvar++)
    t[i] = new_transition_tree_var(d, curvar);
}

/* create a0 = b0 & a1 = b1 & ... & an = bn */
transition_tree equals(trans_data *d, int *a, int *b, int n)
{
  int i;
  transition_tree t;

  t = new_transition_tree_op(d, EQUAL, a[0], b[0]);
  for (i=1; i<n; i++)
    t = new_transition_tree_op(d, AND, t,
	  new_transition_tree_op(d, EQUAL, a[i], b[i]));

  return t;
}

transition_tree equals_mod(trans_data *d, int *to, int nbbits, int start, int *mod, int len)
{
  transition_tree t;
  int i;

  t = new_transition_tree_op(d, EQUAL, to[start], mod[0]);
  for (i=1; i<len; i++)
    t = new_transition_tree_op(d, AND, t,
	  new_transition_tree_op(d, EQUAL, to[start+i], mod[i]));

  for (i=0; i<start; i++)
    t = new_transition_tree_op(d, AND, t,
	  new_transition_tree_op(d, NOT, to[i], TRANS_UNDEF));

  for (i=start+len; i<nbbits; i++)
    t = new_transition_tree_op(d, AND, t,
	  new_transition_tree_op(d, NOT, to[i], TRANS_UNDEF));

  return t;
}

/* returns the tree corresponding to a >= b
 * which is :
 * a0 & !b0 + (a0 = b0) & a(1) >= b(1)
 */
transition_tree test_sup(trans_data *d, int *a, int *b, int n, int i)
{
  transition_tree t, t1;

  if (i==n-1)
    return new_transition_tree_op(d, OR, a[i],
	new_transition_tree_op(d, NOT, b[i], TRANS_UNDEF));

  t = new_transition_tree_op(d, AND, a[i],
	new_transition_tree_op(d, NOT, b[i], TRANS_UNDEF));

  t1 = test_sup(d, a, b, n, i+1);
  t1 = new_transition_tree_op(d, AND, t1,
	new_transition_tree_op(d, EQUAL, a[i], b[i]));

  return new_transition_tree_op(d, OR, t1, t);
}

/* a b carry | a - b | new carry
 * ----------+-------+-----------
 * 0 0   0   |   0   |  0
 * 0 0   1   |   1   |  1
 * 0 1   0   |   1   |  1
 * 0 1   1   |   0   |  1
 * 1 0   0   |   1   |  0
 * 1 0   1   |   0   |  0
 * 1 1   0   |   0   |  0
 * 1 1   1   |   1   |  1
 *
 * c (= a-b) = carry & (a = b) + !carry & (a xor b)
 * nc = !a & (b + carry) + a & b & carry
 */
transition_tree atom_minus(trans_data *d, int a, int b, int c, int carry, int nc)
{
  transition_tree t1, t11, t12, t2, t21, t22;

  t11 = new_transition_tree_op(d, EQUAL, a, b);
  t11 = new_transition_tree_op(d, AND, carry, t11);

  t12 = new_transition_tree_op(d, XOR, a, b);
  t12 = new_transition_tree_op(d, AND,
	  new_transition_tree_op(d, NOT, carry, TRANS_UNDEF),
	  t12);

  t1 = new_transition_tree_op(d, OR, t11, t12);
  t1 = new_transition_tree_op(d, EQUAL, c, t1);

  t21 = new_transition_tree_op(d, OR, b, carry);
  t21 = new_transition_tree_op(d, AND,
	  new_transition_tree_op(d, NOT, a, TRANS_UNDEF),
	  t21);

  t22 = new_transition_tree_op(d, AND, a, b);
  t22 = new_transition_tree_op(d, AND, t22, carry);

  t2 = new_transition_tree_op(d, OR, t21, t22);
  t2 = new_transition_tree_op(d, EQUAL, nc, t2);

  return new_transition_tree_op(d, AND, t1, t2);
}

/* returns :
 * c = a - b
 * ie :
 * a0 a1 .. a(n-1)
 * b0 b1 .. b(n-1)
 * ---------------
 * c0 c1 .. c(n-1)
 * we add a carry, carry[n-1] = 0
 * then we do simple_sub and and it all
 */
transition_tree minus(trans_data *d, int *a, int *b, int *c, int n)
{
  transition_tree t;
  int carry[n+1];
  int i;

  get_variables(d, carry, n+1);

  t = new_transition_tree_op(d, NOT, carry[n], TRANS_UNDEF);

  for (i=n-1; i>=0; i--) {
    t = new_transition_tree_op(d, AND, t,
	  atom_minus(d, a[i], b[i], c[i], carry[i+1], carry[i]));
  }

  return t;
}

/* returns :
 * if a >= b : c = a - b
 * else c = a
 *
 * ie : (a >= b) & (c = a - b) + !(a >= b) & (c = a)
 */
transition_tree strange_sub(trans_data *d, int *a, int *b, int *c, int n)
{
  transition_tree asupb, ceqaminusb, ceqa, notasupb;
  transition_tree pa, pb;

  asupb = test_sup(d, a, b, n, 0);
  notasupb = new_transition_tree_op(d, NOT, asupb, TRANS_UNDEF);

  ceqa = equals(d, c, a, n);
  ceqaminusb = minus(d, a, b, c, n);

  pa = new_transition_tree_op(d, AND, asupb, ceqaminusb);
  pb = new_transition_tree_op(d, AND, notasupb, ceqa);

  return new_transition_tree_op(d, OR, pa, pb);
}

/* returns :
 * a0 a1 ... am ... a(n-1)    divisor
 * b0 b1 ... bm 0 ...    0    var
 * -----------------------
 * c0 c1 ...        c(n-1)    res = divisor
 *  0 b0 b1 ... bm 0 ... 0          var
 * -----------------------
 * ... ...  .. . . . .. .           res
 * etc.
 * at each step :
 *   divisor = previous res (init = num)
 *   var = decal(mod, i)
 *   res = sub(divisor, var)
 * and AND it all
 */
/* nbbits must be >= 16 when coming here and nb >= 2 */
/* nbbits is nb bits of num and nb is nb bits of the modulor (the bit 0 will be forced to 1) */
transition_tree encode(trans_data *d, int *num, int nbbits, int *mod, int nb)
{
  transition_tree t;
  int var[nbbits];
  int divisor[nbbits];
  int res[nbbits];
  int i=0;

  get_variables(d, divisor, nbbits);

  t = equals(d, divisor, num, nbbits);

  do {
    transition_tree t1;

    get_variables(d, var, nbbits);
    t = new_transition_tree_op(d, AND, t, equals_mod(d, var, nbbits, i, mod, nb));

    get_variables(d, res, nbbits);
    t1 = strange_sub(d, divisor, var, res, nbbits);

    t = new_transition_tree_op(d, AND, t, t1);

    /* new divisor = res */
    memcpy(divisor, res, nbbits * sizeof(int));

    i++;
  } while (i<nbbits-nb+1);

#if 1
  /* we want res = 0 */
  for (i=0; i<nbbits; i++)
    t = new_transition_tree_op(d, AND, t,
	  new_transition_tree_op(d, NOT, res[i], TRANS_UNDEF));
#endif

  return t;
}

/* useless, but the prog wants it to compile (because all this is a big rip) */
checker the_checker;

int main(int argc, char **argv)
{
  trans_data d;
  verbose v;
#if 0
  char n[8]={0,0,1,1,0,0,1,0};  /* n = 9 */
  int nv[8];
  int mod[8];
#endif
  char *n;
  int *nv;
  int *mod;
  transition_tree t;
  unsigned long nbvars, nbclauses;
  clauses c;
  int i;
  int factorlen, modlen;

  if (argc != 3)
    goto error;

  factorlen = strlen(argv[1]);

  for (i=0; i<factorlen; i++)
    if (argv[1][i] != '0' && argv[1][i] != '1')
      goto error;

  if (sscanf(argv[2], "%d", &modlen) != 1)
    goto error;

  if (factorlen <= 1 || modlen >= factorlen)
    goto error;

  n = getmem(factorlen * sizeof(char));
  for (i=0; i<factorlen; i++)
    n[i] = argv[1][i] - '0';
  nv = getmem(factorlen * sizeof(int));
  mod = getmem(factorlen * sizeof(int));

  init_verbose(&v);
  //verbose_register(&v, "all");
  init_trans(&d);
  init_clauses(&c);

  get_variables(&d, mod, modlen);
  get_variables(&d, nv, factorlen);

  t = encode(&d, nv, factorlen, mod, modlen);

  for (i=0; i<factorlen; i++) {
    if (n[i])
      t = new_transition_tree_op(&d, AND, t, nv[i]);
    else
      t = new_transition_tree_op(&d, AND, t, new_transition_tree_op(&d, NOT, nv[i], TRANS_UNDEF));
  }

  /* we want the upper bit of divisor set to 1 */
  t = new_transition_tree_op(&d, AND, t, mod[0]);

#if 0
  {
    char modv[8] = {0,0,0,0,0,1,1,1}; /* mod = 5 */
    int res[8];
    get_variables(&d, res, 8);
    t = minus(&d, nv, mod, res, 8);
    for (i=0; i<8; i++) {
      if (n[i])
	t = new_transition_tree_op(&d, AND, t, nv[i]);
      else
	t = new_transition_tree_op(&d, AND, t, new_transition_tree_op(&d, NOT, nv[i], TRANS_UNDEF));

      if (modv[i])
	t = new_transition_tree_op(&d, AND, t, mod[i]);
      else
	t = new_transition_tree_op(&d, AND, t, new_transition_tree_op(&d, NOT, mod[i], TRANS_UNDEF));
    }

  }
#endif

  nbvars = to_cnf(&c, &d, t);
  nbclauses = get_number_of_clauses(&c);

  dump_variable_names(&d, t, 0, 0);
  fprintf(stdout, "p cnf %ld %ld\n", nbvars, nbclauses);
  dump_clauses(&c);
  /* we must have the first equiv set to true otherwise there is always a solution */
  fprintf(stdout, "%d 0\n", d.trans_tree[t].number);

  free_clauses(&c);
  free_trans(&d);
  free_verbose(&v);

  freemem(n);
  freemem(nv);
  freemem(mod);

  return 0;

error:
  fprintf(stderr, "usage : %s <binary number> <divisor length (in bits)>\n", argv[0]);
  fprintf(stderr, "the number must be at least '10' and the divisor length can't be more than number length - 1\n");
  return 1;
}
